import ipaddress as ip
import logging
import os
import typing as t
from pprint import pformat

from goodconf import GoodConf
from pydantic import BaseModel, PositiveInt
from pydantic.types import ImportString

import xpublish

L = logging.getLogger(__name__)


class ClusterConfig(BaseModel):
    module: ImportString
    # Plugin arguments
    args: set[str] = ()
    # Plugin named arguments
    kwargs: dict[str, t.Any] = {}


class PluginConfig(BaseModel):
    module: ImportString
    # Plugin arguments
    args: set[str] = ()
    # Plugin named arguments
    kwargs: dict[str, t.Any] = {}


class RestConfig(GoodConf):
    publish_host: ip.IPv4Address = '0.0.0.0'
    publish_port: PositiveInt = 9000
    log_level: str = 'debug'

    plugins_load_defaults: bool = True
    plugins_config: dict[str, PluginConfig] = {}

    """
    docs_url="/api"
    openapi_url="/api.json"
    """
    app_config: dict[str, t.Any] = {
        'docs_url': '/api',
        'openapi_url': '/api.json',
    }

    """
    available_bytes=1e11
    """
    cache_config: dict[str, t.Any] = {
        'available_bytes': 1e11
    }

    """
    {
        'processes': True,
        'n_workers': 8,
        'threads_per_worker': 1,
        'memory_limit': '4GiB',
    }
    {} = don't load a cluster, parallel=False must be set
    on dataset load or this will cause errors

    None = use default cluster
    """
    cluster_config: ClusterConfig | None = None

    class Config:
        file_env_file = os.environ.get('XPUB_CONFIG_FILE', 'config.yaml')
        env_file = os.environ.get('XPUB_ENV_FILES', '.env')
        env_file_encoding = 'utf-8'
        env_prefix = 'XPUB_'
        env_nested_delimiter = '__'

    def setup_rest(self):

        load_defaults = None
        if self.plugins_load_defaults is False:
            load_defaults = {}

        plugs = self.setup_plugins()

        # Start with no datasets, they are all loaded
        # using the DatasetConfigPlugin
        rest = xpublish.Rest(
            None,
            plugins=load_defaults,
            app_kws=dict(self.app_config),
            cache_kws=dict(self.cache_config),
        )
        for p in plugs.values():
            rest.register_plugin(p, overwrite=True)

        config = {
            'app': dict(self.app_config),
            'cache': dict(self.cache_config),
            'plugins': plugs,
        }
        config_out = pformat(config)
        L.info(config_out)

        return rest

    def setup_cluster(self):
        """
        Load the cluster config we should use for serving this dataset.
        Some integration with Dask Gateway here would be sweet!

        To use no cluster, set to an empty dict. To use the default, set to None.
        This is similar to how plugins work.
        """
        # None or empty - don't load a cluster
        if not self.cluster_config:
            return None

        # Only spawn a cluster if distributed is installed
        try:
            from dask.distributed import Client  # noqa
        except ImportError:
            L.warning("The dask 'distributed' library is not installed, no cluster support")
            return None

        cluster = self.cluster_config.module(
            *self.cluster_config.args,
            **self.cluster_config.kwargs
        )

        L.info(f'Created cluster: {cluster}')
        return cluster

    def setup_plugins(self):
        plugins = {}

        for p in self.plugins_config.values():
            try:
                plug = p.module(
                    *p.args,
                    **p.kwargs
                )
                plugins[plug.name] = plug
            except BaseException as e:
                L.error(f"Could not load the {p} plugin: {e}")

        return plugins

    def setup(self, create_cluster_client=True):
        """
        _summary_

        Args:
            create_cluster_client (bool, optional): When run outside of
                gunicorn this needs to be True for the dask client object
                to be created. Defaults to True.
        """
        if create_cluster_client is True:
            cluster = self.setup_cluster()
            if cluster:
                from dask.distributed import Client
                client = Client(cluster)
                L.info(f'Using cluster: {client.cluster}')
                L.info(f'Dashboard: {client.cluster.dashboard_link}')

        rest = self.setup_rest()
        return rest

    def serve_kwargs(self):
        return dict(
            host=str(self.publish_host),
            port=self.publish_port,
            log_level=self.log_level,
        )



import logging
import os
import re

from fastapi import status
from fastapi.responses import JSONResponse
from uvicorn.workers import UvicornWorker

from xpublish_host.config import RestConfig

logging.basicConfig(level=logging.INFO)
logging.getLogger('distributed').setLevel(logging.ERROR)
logging.getLogger('dask').setLevel(logging.ERROR)

L = logging.getLogger(__name__)


class XpdWorker(UvicornWorker):
    CONFIG_KWARGS = {
        "factory": True,
    }


def health_check(request):
    return JSONResponse(
        {'xpublish': 'online'},
        status_code=status.HTTP_200_OK,
        media_type="application/health+json",
    )


def setup_health(app):
    if os.environ.get("XPUB_HEALTH_DISABLE"):
        return

    health = os.environ.get("XPUB_HEALTH_ENDPOINT", "/health")

    app.add_route(health, health_check)
    return health


def get_dataset_label(request):
    try:
        pattern = r'^.*\/datasets\/(\w+)\/.*$'
        return re.match(pattern, str(request.url)).group(1)
    except BaseException:
        return ''


def setup_metrics(app, health_endpoint):

    if os.environ.get("XPUB_METRICS_DISABLE"):
        return

    envvar = 'PROMETHEUS_MULTIPROC_DIR'
    if envvar not in os.environ:
        L.warning(f"{envvar} is not set! Metrics will not work if using gunicorn.")

    app_name = os.environ.get("XPUB_METRICS_APP_NAME", "xpublish")
    prefix_name = os.environ.get("XPUB_METRICS_PREFIX_NAME", "xpublish_host")
    environment = os.environ.get("XPUB_METRICS_ENVIRONMENT", "development")
    metrics = os.environ.get("XPUB_METRICS_ENDPOINT", "/metrics")

    try:
        from starlette_exporter import PrometheusMiddleware, handle_metrics
        from starlette_exporter.optional_metrics import request_body_size, response_body_size

        app.add_middleware(
            PrometheusMiddleware,
            app_name=app_name,
            prefix=prefix_name,
            buckets=[0.01, 0.1, 0.25, 0.5, 1.0],
            skip_paths=[health_endpoint, metrics, '/favicon.ico'],
            group_paths=False,
            optional_metrics=[response_body_size, request_body_size],
            labels=dict(
                environment=environment,
                dataset=get_dataset_label,
            )
        )
        app.add_route(metrics, handle_metrics)
    except BaseException:
        raise


# def setup_config(config_file: str = None, **setup_kwargs):
#     env_config = os.environ.get('XPUB_ENV_FILES', None)

#     # Load in any passed in config file, or use ENV variables
#     # that override any defined in the env file
#     if config_file:
#         config = RestConfig(_env_file=env_config, load=False)
#         config.load(config_file)
#         return config

#     # Look for environmental variable defining the location of a config file
#     config_file = os.environ.get('XPUB_CONFIG_FILE', None)
#     if config_file and os.path.exists(config_file):
#         config = RestConfig(_env_file=env_config, load=False)
#         config.load(config_file)
#         return config

#     return RestConfig(_env_file=env_config)


def setup_config(config_file: str = None, **setup_kwargs):
    env_config = os.environ.get('XPUB_ENV_FILES', None)
    config = RestConfig(_env_file=env_config)

    # Look for environmental variable defining the location
    # of a config file
    yaml_config = os.environ.get('XPUB_CONFIG_FILE', None)
    if yaml_config and os.path.exists(yaml_config):
        config.load(yaml_config)

    # Load in any passed in config file, or use ENV variables
    # that override any defined in the env file
    if config_file:
        config.load(config_file)

    if not yaml_config and not config_file:
        config.load()

    return config


def setup_xpublish(config: RestConfig = None, **setup_kwargs):
    rest = config.setup(**setup_kwargs)
    app = rest.app
    health_endpoint = setup_health(app)
    _ = setup_metrics(app, health_endpoint)
    rest._app = app

    return rest, config


def serve(config_file: str = None):
    config = setup_config(config_file)
    rest, config = setup_xpublish(config)
    rest.serve(
        **config.serve_kwargs()
    )


def app():
    config = setup_config()
    rest, _ = setup_xpublish(config, create_cluster_client=False)
    return rest.app


def run():
    import argparse
    parser = argparse.ArgumentParser("xpublish_host_serve")
    parser.add_argument("-c", "--config", default=None, help="Path to a config file", type=str)
    args = parser.parse_args()

    if args.config and not os.path.exists(args.config):
        raise ValueError(f"File {args.config} not found")
    serve(config_file=args.config)


if __name__ == '__main__':
    run()



import itertools
import types
import weakref
from typing import (
    Generic,
    List,
    TypeVar,
)


class TypeParametersMemoizer(type):
    """
    https://github.com/tiangolo/fastapi/discussions/8225#discussioncomment-5149945
    """
    _generics_cache = weakref.WeakValueDictionary()

    def __getitem__(cls, typeparams):

        # prevent duplication of generic types
        if typeparams in cls._generics_cache:
            return cls._generics_cache[typeparams]

        # middleware class for holding type parameters
        class TypeParamsWrapper(cls):
            __type_parameters__ = typeparams if isinstance(typeparams, tuple) else (typeparams,)

            @classmethod
            def _get_type_parameters(cls):
                return cls.__type_parameters__

        return types.GenericAlias(TypeParamsWrapper, typeparams)


T = TypeVar('T')


class CommaSeparatedList(Generic[T]):
    """
    A custom type for comma-separated lists compatible with Pydantic.
    """
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: str | List[str], values=None, field=None):
        if isinstance(v, str):
            v = v.split(",")
        else:
            v = list(itertools.chain.from_iterable((x.split(",") for x in v)))
        v = list(map(str.strip, v))
        if len(v) < 1:
            raise ValueError("List must contain at least one item")
        return v



import logging
import typing as t
from operator import attrgetter
from pathlib import Path

import xarray as xr

L = logging.getLogger(__name__)


def load_mfdataset(
    root_path: str | Path,
    file_glob: str,
    open_mfdataset_kwargs: t.Dict = {},
    file_limit: int | None = None,
    skip_head_files: int | None = 0,
    skip_tail_files: int | None = 0,
    computes: list[str] | None = None,
    chunks: dict[str, int] | None = None,
    axes: dict[str, str] | None = None,
    sort_by: dict[str, str] | None = None,
    isel: dict[str, slice] | None = None,
    sel: dict[str, slice] | None = None,
    rechunk: bool = False,
    attrs_file_idx: int = -1,
    combine_by_coords: list[str | Path] = None,
    **kwargs
) -> xr.Dataset:

    # drops = drops or []
    computes = computes or []
    chunks = chunks or {}
    axes = axes or {}
    sort_by = sort_by or {}
    isel = isel or {}
    sel = sel or {}
    combine_by_coords = combine_by_coords or []

    root = Path(root_path)
    files = sorted(
        [
            p for p in root.glob(file_glob)
        ],
        key=attrgetter('name')
    )

    # Skip files from the front and back. If not defined
    # (None) this won't change the files list.
    if skip_tail_files:
        skip_tail_files = skip_tail_files * -1
    else:
        # Prevents a zero from not working as expected
        skip_tail_files = None

    files = files[skip_head_files:skip_tail_files]

    # You know, for testing
    if file_limit:
        files = files[-file_limit:]

    num_files = len(files)
    L.info(f"Found {num_files} files in {root_path}{file_glob}")
    if not num_files:
        return xr.Dataset()

    # Set a default chunking scheme if one was not provided
    # that uses all defined axes chunked by 'auto'
    axis_names = ['t', 'z', 'x', 'y']
    if not chunks:
        chunks = {
            axes.get(a, a): 'auto'
            for a in axis_names
            if a in axes and not chunks
        }
    L.info(f'Using chunking scheme: {chunks}')

    # Pull metadata from the last file unless a file was specified
    attrs_file = open_mfdataset_kwargs.pop('attrs_file', files[attrs_file_idx])

    # These are the default open_mfdataset_kwargs
    xr_kwargs = dict(
        parallel=True,
        engine='netcdf4',
        data_vars='minimal',
        coords='minimal',
        compat='override',
        combine='nested',
        concat_dim=[
            axes.get('t', 't')
        ],
        decode_cf=True,
        decode_times=True,
        decode_timedelta=False,
        chunks=chunks,
        attrs_file=attrs_file,
    )
    xr_kwargs.update(open_mfdataset_kwargs)

    L.info(f"Loading {num_files} files with {xr_kwargs}...")
    cache_size = max(num_files, 128)
    xr.set_options(file_cache_maxsize=cache_size)
    ds = xr.open_mfdataset(
        files,
        **xr_kwargs
    )

    if combine_by_coords:
        for combine_file in combine_by_coords:
            L.info(f"Combining {combine_file}...")
            combo = xr.open_dataset(
                combine_file,
                engine='netcdf4',
                drop_variables=xr_kwargs.get('drop_variables', [])
            )
            ds = xr.combine_by_coords(
                [combo, ds],
                compat='override',
                combine_attrs='override'
            )

    if sort_by:
        L.info(f"Sorting by {sort_by}...")
        ds = ds.sortby(sort_by)

    if isel:
        L.info("Selecting by index...")
        isels = {
            k: slice(*v) for k, v in isel.items()
        }
        ds = ds.isel(**isels)

    if sel:
        L.info("Selecting...")
        ds = ds.sel(**sel)

    if rechunk is True and chunks:
        L.info("Rechunking...")
        # Remove any dims that may have been squashed due to processing
        chunks = { k: v for k, v in chunks.items() if k in ds.dims }
        ds = ds.chunk(chunks)

    L.info("Computing and assigning axes as coordinates...")
    assigns = {}
    for _, aname in axes.items():
        assigns[aname] = ds[aname].compute()
    ds = ds.assign_coords(assigns)

    L.info("Computing variables...")
    for variable in computes:
        ds[variable] = ds[variable].compute()

    return ds



import os

APP_NAME = os.environ.get("XPUB_METRICS_APP_NAME", "xpublish")
PREFIX_NAME = os.environ.get("XPUB_METRICS_PREFIX_NAME", "xpublish_host")
ENVIRONMENT = os.environ.get("XPUB_METRICS_ENVIRONMENT", "development")

DEFAULT_LABELS = {
    'environment': ENVIRONMENT,
    'app_name': APP_NAME,
}


def create_metric(metric_type, name, description, labels, *args, **kwargs):
    labels += list(DEFAULT_LABELS.keys())
    return metric_type(
        f"{PREFIX_NAME}_{name}",
        description,
        labels,
        *args,
        **kwargs
    )



from prometheus_client import multiprocess

proc_name = 'xpublish'
bind = "0.0.0.0:9000"
workers = 4
threads = 1
worker_class = 'xpublish_host.app.XpdWorker'
accesslog = '-'
errorlog = '-'
loglevel = 'info'
capture_output = True
preload_app = True
timeout = 0
keepalive = 10
graceful_timeout = 10


def child_exit(server, worker):
    try:
        multiprocess.mark_process_dead(worker.pid)
    except TypeError:
        pass


def on_starting(server):
    """
    Create a dask cluster object as needed and store the
    scheduler address for later usage in the worker process
    init
    """
    from xpublish_host.app import setup_config
    config = setup_config()
    cluster = config.setup_cluster()
    if hasattr(cluster, 'scheduler_address'):
        server.XPUB_DASK_SCHEDULER_ADDRESS = cluster.scheduler_address


def post_fork(server, worker):
    """
    In each worker, connect to the scheduler address of the dask cluster
    that has already been configured.
    """
    if hasattr(server, 'XPUB_DASK_SCHEDULER_ADDRESS'):
        from dask.distributed import Client
        client = Client(server.XPUB_DASK_SCHEDULER_ADDRESS)
        print(f'Worker {worker.pid} is using cluster: {client} at {client.dashboard_link}')



import xarray as xr


def simple():
    return xr.Dataset(
        {
            'count': ('x', [1, 2, 3]),
            'scalar': 1
        }
    )


def kwargs(varname, values=None):
    return xr.Dataset({varname: ('x', values)})



import logging
from pathlib import Path

import xarray as xr

L = logging.getLogger(__name__)


def load_dataset_zarr(json_path: str | Path, chunks=None):

    # do this to 1. use dask and 2. ensure chunks match between
    # encoding and dask arrays
    # https://github.com/xpublish-community/xpublish/issues/207
    chunks = chunks or {}

    L.info(f"Chunks are {chunks}.")

    L.info(f"using json from {json_path}")

    ds = xr.open_dataset(
        "reference://", engine="zarr",
        backend_kwargs={
            "storage_options": {
                "fo": json_path,
            },
            "consolidated": False
        },
        chunks=chunks,
    )

    # remove chunks encoding, same issue as above
    # https://github.com/xpublish-community/xpublish/issues/207
    L.info("removing chunks from all variable encoding...")
    for varname in ds.variables:
        if "chunks" in ds[varname].encoding:
            L.info(f"removing chunks from {varname}")
            del ds[varname].encoding["chunks"]

    # ds = ds.reset_encoding()
    return ds



import logging
import os
import typing as t
from datetime import datetime, timezone

import xarray as xr
from goodconf import GoodConf
from pydantic import BaseModel, FilePath
from pydantic.types import ImportString

from xpublish import Plugin, hookimpl
from xpublish_host.config import RestConfig

try:
    from prometheus_client import Counter, Gauge

    from xpublish_host.metrics import DEFAULT_LABELS, create_metric
    metrics = True
    DATASET_LOAD_TIME = create_metric(
        Gauge,
        "dataset_load_time",
        "How long it look to last load the dataset",
        ["dataset"],
    )
    DATASET_LOAD_COUNT = create_metric(
        Counter,
        "dataset_load_count",
        "How many times a dataset has been loaded",
        ["dataset"],
    )
    DATASET_LOAD_WHEN = create_metric(
        Gauge,
        "dataset_load_when",
        "When the dataset was last loaded",
        ["dataset"],
    )
except ImportError:
    metrics = False

L = logging.getLogger(__name__)


class DatasetConfig(BaseModel):
    id: str
    title: str
    description: str
    loader: ImportString
    args: list[t.Any] = ()
    kwargs: dict[str, t.Any] = {}
    invalidate_after: int | None = None
    skip_initial_load: bool = False

    def load(self):
        return self.loader(*self.args, **self.kwargs)

    def serve(self, **rest_kwargs):
        """
        Helper method to run a single dataset with configs
        """
        config = RestConfig(**rest_kwargs, load=True)
        rest = config.setup()
        rest.register_plugin(
            DatasetsConfigPlugin(
                dataset_config={self.id: self}
            ),
            overwrite=True
        )
        rest.serve(
            **config.serve_kwargs()
        )


class DatasetConfigFile(GoodConf):
    datasets_config: dict[str, DatasetConfig] = {}


class DatasetsConfigPlugin(Plugin):

    name: t.ClassVar[str] = 'dconfig'

    datasets_config: dict[str, DatasetConfig] = {}
    datasets_config_file: FilePath = None

    __datasets: dict = {}
    __datasets_loaded: dict = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        config_file_datasets = self.load_config_file()

        self.datasets_config.update(config_file_datasets)

        for dsc in self.datasets_config.values():
            if dsc.skip_initial_load is False:
                L.info(f"Loading dataset (initial): {dsc.id}")
                _ = self.load_dataset(dsc)

    def load_config_file(self):
        # Load a config file into a dict of DatasetConfigs
        dcf = DatasetConfigFile(load=True)

        # Look for env file pointing to a config file
        yaml_config = os.environ.get('XPUBDC_CONFIG_FILE', None)
        if yaml_config and os.path.exists(yaml_config):
            dcf.load(yaml_config)

        if self.datasets_config_file and os.path.exists(self.datasets_config_file):
            dcf.load(self.datasets_config_file)

        if dcf.datasets_config:
            return dcf.datasets_config

        return {}

    @hookimpl
    def get_datasets(self):
        return [ v.id for v in self.datasets_config.values() ]

    @hookimpl
    def get_dataset(self, dataset_id: str) -> xr.Dataset:

        try:
            dsc = next(v for v in self.datasets_config.values() if v.id == dataset_id)
        except StopIteration:
            return

        # TODO: Cache check. We could potentially check a cache to see if this dataset
        # should be reloaded after a certain timeout or key expiration. This could be
        # set through an HTTP endpoint to invalidate a dataset's cache. This cache needs
        # to be on the per-process level since that is where the datasets are accessed
        cache_check = True

        # Expiration check, after X amount of time, invalidate the dataset
        expiration_check = True
        if dsc.invalidate_after is not None:
            last_updated = self.__datasets_loaded.get(dataset_id, 0)
            now = datetime.now(timezone.utc).timestamp()
            expiration_check = (now - last_updated) < dsc.invalidate_after

        if dataset_id in self.__datasets and cache_check and expiration_check:
            return self.__datasets[dataset_id]

        # If we got here, load the dataset
        L.info(f"Loading dataset: {dsc.id}")
        dataset = self.load_dataset(dsc)
        return dataset

    def load_dataset(self, config: DatasetConfig):
        now = datetime.utcnow().timestamp()
        dataset = config.load()

        if metrics is True:
            after = datetime.utcnow().timestamp()
            elapsed = after - now
            DATASET_LOAD_TIME.labels(dataset=config.id, **DEFAULT_LABELS).set(elapsed)
            DATASET_LOAD_WHEN.labels(dataset=config.id, **DEFAULT_LABELS).set(after)
            DATASET_LOAD_COUNT.labels(dataset=config.id, **DEFAULT_LABELS).inc()

        self.__datasets[config.id] = dataset
        self.__datasets_loaded[config.id] = now
        return self.__datasets[config.id]



import io
import logging
from datetime import datetime, timezone
from enum import Enum
from typing import Annotated, Sequence

from fastapi import (
    APIRouter,
    Depends,
    Query,
    Response,
)
from fastapi.responses import StreamingResponse

from xpublish.plugins import (
    Dependencies,
    Plugin,
    hookimpl,
)
from xpublish_host.utils import CommaSeparatedList

L = logging.getLogger(__name__)


def utc_native_dt(dt):
    if isinstance(dt, datetime):
        if dt.tzinfo:
            dt = dt.astimezone(timezone.utc).replace(tzinfo=None)
    return dt


class DataFormat(str, Enum):
    JSONL = ".jsonl"
    DICT = '.dict'
    LIST = '.list'
    SPLIT = '.split'
    TIGHT = '.tight'
    RECORDS = '.records'
    INDEX = '.index'
    PARQUET = '.parquet'


class DataPointsPlugin(Plugin):
    """Adds an Data Point Extraction endpoint"""

    name: str = 'data_points'  # Add type annotation

    dataset_router_prefix: str = '/data_points'
    dataset_router_tags: Sequence[str] = ['data_points']

    @hookimpl
    def dataset_router(self, deps: Dependencies):

        def grid_params(
            x_var: Annotated[str | None, Query()] = None,
            y_var: Annotated[str | None, Query()] = None,
        ):
            return {
                'x_var': x_var,
                'y_var': y_var,
            }

        def depth_params(
            depth_var: Annotated[str | None, Query()] = None,
            depth_start: Annotated[float | None, Query()] = 0,
            depth_end: Annotated[float | None, Query()] = 1,
        ):
            return {
                'var': depth_var,
                'start': depth_start,
                'end': depth_end,
            }

        def time_params(
            time_var: Annotated[str | None, Query()] = None,
            time_start: Annotated[datetime | None, Query()] = None,
            time_end: Annotated[datetime | None, Query()] = None,
        ):
            return {
                'var': time_var,
                'start': utc_native_dt(time_start),
                'end': utc_native_dt(time_end)
            }

        def var_params(
            var: Annotated[CommaSeparatedList[str] | None, Query()] = None,
            keep: Annotated[CommaSeparatedList[str] | None, Query()] = None,
            return_null: Annotated[bool, Query()] = False,
        ):
            return {
                'var': var,
                'keep': keep,
                'return_null': return_null,
            }

        router = APIRouter(prefix=self.dataset_router_prefix, tags=list(self.dataset_router_tags))

        @router.get('/filter{fmt}', summary="Gets data points between 2 times for a list of variables")
        async def get_points(
            fmt: DataFormat = '.jsonl',
            dataset=Depends(deps.dataset),
            time_params=Depends(time_params),
            depth_params=Depends(depth_params),
            var_params=Depends(var_params),
            grid_params=Depends(grid_params),
        ):

            selection = {}
            renames = {}

            if time_params['var']:

                renames[time_params['var']] = 't'

                if time_params['var'] in dataset.dims:
                    selection[time_params['var']] = slice(
                        time_params['start'],
                        time_params['end']
                    )
                else:
                    L.warning(f"'{ time_params['var']}' not found in dataset dimensions")

            if depth_params['var']:

                renames[depth_params['var']] = 'z'

                if depth_params['var'] in dataset.dims:
                    selection[depth_params['var']] = slice(
                        depth_params['start'],
                        depth_params['end']
                    )
                else:
                    L.warning(f"'{ depth_params['var']}' not found in dataset dimensions")

            if grid_params['x_var']:
                renames[grid_params['x_var']] = 'x'

            if grid_params['y_var']:
                renames[grid_params['y_var']] = 'y'

            # How far back to return data for
            ds = dataset.sel(selection)

            # Subset to requested variables
            if var_params['var']:
                ds = ds[var_params['var']]

            # Convert to a dataframe
            df = ds.to_dataframe().reset_index()

            axis_vars = [
                renames.get(time_params['var'], None),
                renames.get(depth_params['var'], None),
                renames.get(grid_params['x_var'], None),
                renames.get(grid_params['y_var'], None),
            ]
            axis_vars = [ x for x in axis_vars if x ]

            keep = axis_vars.copy()
            keep += var_params['var'] or []
            keep += var_params['keep'] or []

            if var_params['return_null'] is False:
                df = df.dropna(how='all', subset=var_params['var'])

            df = df.reset_index(drop=True)

            # standardize axes
            if renames:
                df = df.rename(columns=renames)

            df = df.drop(columns=[ c for c in df.columns if c not in keep ])

            if fmt == DataFormat.JSONL:
                data = df.to_json(
                    orient='records',
                    lines=True,
                    date_format='iso',
                )
                return Response(
                    data,
                    media_type='application/jsonlines+json'
                )
            elif fmt == DataFormat.PARQUET:

                data = io.BytesIO()
                df.to_parquet(
                    data,
                    index=None,
                )

                def stream():
                    data.seek(0)
                    yield from data

                return StreamingResponse(
                    stream(),
                    # https://issues.apache.org/jira/browse/PARQUET-1889
                    media_type="application/vnd.apache.parquet"
                )
            else:
                if fmt in [
                    DataFormat.SPLIT,
                    DataFormat.TIGHT,
                ]:
                    if axis_vars:
                        df = df.set_index(axis_vars)

                data = df.to_dict(
                    orient=fmt[1:]  # strip out the leading period,
                )
                return data

        return router
